'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var backendCommon = require('@backstage/backend-common');
var express = require('express');
var Router = require('express-promise-router');
var expressGraphql = require('express-graphql');
var graphql = require('graphql');
var core = require('@envelop/core');
var graphqlModules$1 = require('@envelop/graphql-modules');
var graphqlModules = require('graphql-modules');
var schema = require('@graphql-tools/schema');
var loadFiles = require('@graphql-tools/load-files');
var GraphQLJSON = require('graphql-type-json');
var lodash = require('lodash');
var graphqlRelay = require('graphql-relay');
var catalogModel = require('@backstage/catalog-model');
var utils = require('@graphql-tools/utils');
var dataloader = require('@envelop/dataloader');
var DataLoader = require('dataloader');
var merge = require('@graphql-tools/merge');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var express__default = /*#__PURE__*/_interopDefaultLegacy(express);
var Router__default = /*#__PURE__*/_interopDefaultLegacy(Router);
var GraphQLJSON__default = /*#__PURE__*/_interopDefaultLegacy(GraphQLJSON);
var DataLoader__default = /*#__PURE__*/_interopDefaultLegacy(DataLoader);

const Catalog = graphqlModules.createModule({
  id: "catalog",
  typeDefs: loadFiles.loadFilesSync(backendCommon.resolvePackagePath("@frontside/backstage-plugin-graphql", "src/app/modules/catalog/catalog.graphql")),
  resolvers: {
    Lifecycle: {
      EXPERIMENTAL: "experimental",
      PRODUCTION: "production",
      DEPRECATED: "deprecated"
    },
    JSON: GraphQLJSON__default["default"],
    JSONObject: GraphQLJSON.GraphQLJSONObject,
    Query: {
      entity: (_, { name, kind, namespace }, { refToId }) => ({ id: refToId({ name, kind, namespace }) })
    }
  }
});

const Core = graphqlModules.createModule({
  id: "core",
  typeDefs: loadFiles.loadFilesSync(backendCommon.resolvePackagePath("@frontside/backstage-plugin-graphql", "src/app/modules/core/core.graphql")),
  resolvers: {
    Node: {
      id: async ({ id }, _, { loader }) => {
        const entity = await loader.load(id);
        if (!entity)
          return null;
        return id;
      }
    },
    Query: { node: (_, { id }) => ({ id }) }
  }
});

function getObjectTypeName(iface, directive) {
  if (directive && "generatedTypeName" in directive)
    return directive.generatedTypeName;
  return iface.name.slice(1);
}
function filterEntityRefs(entity, relationType, targetKind) {
  var _a, _b;
  return (_b = (_a = entity == null ? void 0 : entity.relations) == null ? void 0 : _a.filter(({ type }) => !relationType || type === relationType).flatMap(({ targetRef }) => {
    const ref = catalogModel.parseEntityRef(targetRef);
    return !targetKind || ref.kind.toLowerCase() === targetKind.toLowerCase() ? [ref] : [];
  })) != null ? _b : [];
}
function isConnectionType(type) {
  return graphql.isInterfaceType(type) && type.name === "Connection" || graphql.isNonNullType(type) && isConnectionType(type.ofType);
}
function createConnectionType(typeName, fieldType, nodeType) {
  const wrappedEdgeType = fieldType.getFields().edges.type;
  const edgeType = wrappedEdgeType.ofType.ofType.ofType;
  return new graphql.GraphQLObjectType({
    name: `${typeName}Connection`,
    fields: {
      ...fieldType.toConfig().fields,
      edges: {
        type: new graphql.GraphQLNonNull(new graphql.GraphQLList(new graphql.GraphQLNonNull(new graphql.GraphQLObjectType({
          name: `${typeName}Edge`,
          fields: {
            ...edgeType.toConfig().fields,
            node: {
              type: new graphql.GraphQLNonNull(nodeType)
            }
          },
          interfaces: [edgeType]
        }))))
      }
    },
    interfaces: [fieldType]
  });
}
function transformDirectives(sourceSchema) {
  const extendsWithoutArgs = /* @__PURE__ */ new Set();
  const resolversMap = {};
  const typesToAdd = /* @__PURE__ */ new Map();
  const additionalInterfaces = {};
  function handleFieldDirective(field, directive) {
    if ("at" in directive && typeof directive.at !== "string" && (!Array.isArray(directive.at) || directive.at.some((a) => typeof a !== "string"))) {
      throw new Error(`The "at" argument of @field directive must be a string or an array of strings`);
    }
    field.resolve = async ({ id }, _, { loader }) => {
      var _a;
      const entity = await loader.load(id);
      if (!entity)
        return null;
      return (_a = lodash.get(entity, directive.at)) != null ? _a : directive.default;
    };
  }
  function handleRelationDirective(field, directive, schema) {
    var _a;
    const fieldType = field.type;
    if (graphql.isListType(fieldType) && isConnectionType(fieldType.ofType) || graphql.isNonNullType(fieldType) && graphql.isListType(fieldType.ofType) && isConnectionType(fieldType.ofType.ofType)) {
      throw new Error(`It's not possible to use a list of Connection type. Use either Connection type or list of specific type`);
    }
    const isList = graphql.isListType(fieldType) || graphql.isNonNullType(fieldType) && graphql.isListType(fieldType.ofType);
    if (isConnectionType(fieldType)) {
      if (directive.nodeType) {
        const nodeType = schema.getType(directive.nodeType);
        if (!nodeType) {
          throw new Error(`The interface "${directive.nodeType}" is not defined in the schema.`);
        }
        if (graphql.isInputType(nodeType)) {
          throw new Error(`The interface "${directive.nodeType}" is an input type and can't be used in a Connection.`);
        }
        if (graphql.isUnionType(nodeType)) {
          const iface = (_a = typesToAdd.get(directive.nodeType)) != null ? _a : new graphql.GraphQLInterfaceType({
            name: directive.nodeType,
            interfaces: [schema.getType("Node")],
            fields: { id: { type: new graphql.GraphQLNonNull(graphql.GraphQLID) } },
            resolveType: (...args2) => resolversMap.Node(...args2)
          });
          typesToAdd.set(directive.nodeType, iface);
          nodeType.getTypes().forEach((type) => {
            var _a2;
            additionalInterfaces[type.name] = ((_a2 = additionalInterfaces[type.name]) != null ? _a2 : /* @__PURE__ */ new Set()).add(iface);
          });
          field.type = createConnectionType(nodeType.name, fieldType, iface);
        } else {
          const typeName = graphql.isInterfaceType(nodeType) ? getObjectTypeName(nodeType) : nodeType.name;
          field.type = createConnectionType(typeName, fieldType, nodeType);
        }
      }
      const mandatoryArgs = [
        ["first", "Int"],
        ["after", "String"],
        ["last", "Int"],
        ["before", "String"]
      ];
      const args = { ...field.args };
      mandatoryArgs.forEach(([name, type]) => {
        if (name in args) {
          const argType = args[name].type;
          if ((graphql.isNonNullType(argType) ? argType.ofType.toString() : argType.name) !== type) {
            throw new Error(`The field has mandatory argument "${name}" with different type than expected. Expected: ${type}`);
          }
        }
        args[name] = { type: type === "Int" ? graphql.GraphQLInt : graphql.GraphQLString };
      });
      field.args = args;
      field.resolve = async ({ id }, args2, { loader, refToId }) => {
        const ids = filterEntityRefs(await loader.load(id), directive.name, directive.kind).map((ref) => ({ id: refToId(ref) }));
        return {
          ...graphqlRelay.connectionFromArray(ids, args2),
          count: ids.length
        };
      };
    } else {
      field.resolve = async ({ id }, _, { loader, refToId }) => {
        var _a2;
        const ids = filterEntityRefs(await loader.load(id), directive.name, directive.kind).map((ref) => ({ id: refToId(ref) }));
        return isList ? ids : (_a2 = ids[0]) != null ? _a2 : null;
      };
    }
  }
  function validateExtendDirective(interfaceType, directive, schema) {
    var _a;
    if (!/^I[A-Z].*/.test(interfaceType.name)) {
      if (!("generatedTypeName" in directive)) {
        throw new Error(`The interface name "${interfaceType.name}" should started from capitalized 'I', like: "I${interfaceType.name}"`);
      }
      if (schema.getType(directive.generatedTypeName)) {
        throw new Error(`The type "${directive.generatedTypeName}" described in the @extend directive is already declared in the schema`);
      }
    }
    if ("when" in directive !== "is" in directive) {
      throw new Error(`The @extend directive for "${interfaceType.name}" should have both "when" and "is" arguments or none of them`);
    }
    if (!("when" in directive)) {
      if ("interface" in directive && extendsWithoutArgs.has(directive.interface)) {
        throw new Error(`The @extend directive of "${directive.interface}" without "when" and "is" arguments could be used only once`);
      } else {
        extendsWithoutArgs.add(directive.interface);
      }
      const parentType = schema.getType(directive.interface);
      if (parentType) {
        const [extendDirective] = (_a = utils.getDirective(schema, parentType, "extend")) != null ? _a : [];
        if (extendDirective && "interface" in extendDirective && !("when" in extendDirective) && Object.values(interfaceType.getFields()).some((field) => graphql.isNonNullType(field.type))) {
          throw new Error(`The interface "${interfaceType.name}" has required fields and can't be extended from "${directive.interface}" without "when" and "is" arguments, because "${directive.interface}" has already been extended without them`);
        }
      }
    }
    if ("when" in directive && (typeof directive.when !== "string" || Array.isArray(directive.when) && directive.when.some((a) => typeof a !== "string"))) {
      throw new Error(`The "when" argument of @extend directive must be a string or an array of strings`);
    }
  }
  function defineResolver(iface, directive, schema) {
    var _a, _b;
    const objectType = getObjectTypeName(iface, directive);
    if (!resolversMap[iface.name])
      resolversMap[iface.name] = () => objectType;
    const extendType = schema.getType(directive.interface);
    if (!extendType)
      return;
    const [extendDirective] = (_a = utils.getDirective(schema, extendType, "extend")) != null ? _a : [];
    const extendedObjectType = getObjectTypeName(extendType, extendDirective);
    const resolveType = (_b = resolversMap[extendType.name]) != null ? _b : extendType.name === "Node" ? () => void 0 : () => extendedObjectType;
    resolversMap[extendType.name] = async (source, context, info, abstractType) => {
      var _a2, _b2, _c, _d, _e;
      if ("when" in directive && "is" in directive) {
        const { id } = source;
        const { loader } = context;
        const entity = await loader.load(id);
        if (!entity)
          return void 0;
        if (lodash.isEqual(lodash.get(entity, directive.when), directive.is)) {
          return (_b2 = (_a2 = resolversMap[iface.name]) == null ? void 0 : _a2.call(resolversMap, source, context, info, abstractType)) != null ? _b2 : void 0;
        }
        return (_c = resolveType(source, context, info, abstractType)) != null ? _c : void 0;
      }
      return (_e = (_d = resolversMap[iface.name]) == null ? void 0 : _d.call(resolversMap, source, context, info, abstractType)) != null ? _e : void 0;
    };
  }
  const finalSchema = utils.mapSchema(utils.addTypes(utils.mapSchema(utils.mapSchema(sourceSchema, {
    [utils.MapperKind.COMPOSITE_FIELD]: (fieldConfig, fieldName, typeName, schema) => {
      var _a, _b;
      const [fieldDirective] = (_a = utils.getDirective(schema, fieldConfig, "field")) != null ? _a : [];
      const [relationDirective] = (_b = utils.getDirective(schema, fieldConfig, "relation")) != null ? _b : [];
      if (fieldDirective && relationDirective) {
        throw new Error(`The field "${fieldName}" of "${typeName}" type has both @field and @relation directives at the same time`);
      }
      try {
        if (fieldDirective) {
          handleFieldDirective(fieldConfig, fieldDirective);
        } else if (relationDirective) {
          handleRelationDirective(fieldConfig, relationDirective, schema);
        }
      } catch (error) {
        const errorMessage = error instanceof Error ? error.message : error;
        throw new Error(`Error while processing directives on field "${fieldName}" of "${typeName}":
${errorMessage}`);
      }
      return fieldConfig;
    }
  }), {
    [utils.MapperKind.INTERFACE_TYPE]: (interfaceType, schema) => {
      var _a, _b, _c;
      if (interfaceType.name === "Node") {
        interfaceType.resolveType = (...args) => resolversMap[interfaceType.name](...args);
      }
      const [extendDirective] = (_a = utils.getDirective(schema, interfaceType, "extend")) != null ? _a : [];
      if (!extendDirective)
        return interfaceType;
      validateExtendDirective(interfaceType, extendDirective, schema);
      defineResolver(interfaceType, extendDirective, schema);
      const objectType = getObjectTypeName(interfaceType, extendDirective);
      const extendInterfaces = traverseExtends(interfaceType, schema);
      const interfaces = [...new Map([
        ...(_c = (_b = additionalInterfaces[interfaceType.name]) == null ? void 0 : _b.values()) != null ? _c : [],
        ...extendInterfaces.flatMap((iface) => {
          var _a2, _b2;
          return [...(_b2 = (_a2 = additionalInterfaces[iface.name]) == null ? void 0 : _a2.values()) != null ? _b2 : []];
        }),
        ...extendInterfaces
      ].map((iface) => [iface.name, iface])).values()];
      const fields = [...interfaces].reverse().reduce((acc, type) => ({ ...acc, ...type.toConfig().fields }), {});
      const { astNode, extensionASTNodes, ...typeConfig } = interfaceType.toConfig();
      typesToAdd.set(objectType, new graphql.GraphQLObjectType({ ...typeConfig, name: objectType, fields, interfaces }));
      return new graphql.GraphQLInterfaceType({
        ...typeConfig,
        fields,
        resolveType: (...args) => resolversMap[interfaceType.name](...args),
        interfaces: interfaces.filter((iface) => iface.name !== interfaceType.name)
      });
    }
  }), [...typesToAdd.values()]), {
    [utils.MapperKind.UNION_TYPE]: (unionType, schema) => {
      const typeConfig = unionType.toConfig();
      if (!typeConfig.types.some((type) => graphql.isInterfaceType(type) && type.name in resolversMap))
        return unionType;
      typeConfig.types = typeConfig.types.flatMap((type) => {
        if (graphql.isInterfaceType(type)) {
          return utils.getImplementingTypes(type.name, schema).map((name) => schema.getType(name));
        }
        return [type];
      });
      typeConfig.resolveType = (...args) => resolversMap.Node(...args);
      return new graphql.GraphQLUnionType(typeConfig);
    }
  });
  return finalSchema;
}
function traverseExtends(type, schema) {
  var _a;
  const [extendDirective] = (_a = utils.getDirective(schema, type, "extend")) != null ? _a : [];
  const interfaces = [type, ...type.getInterfaces().flatMap((iface) => traverseExtends(iface, schema))];
  if (extendDirective && "interface" in extendDirective) {
    const extendType = schema.getType(extendDirective.interface);
    if (!graphql.isInterfaceType(extendType)) {
      throw new Error(`The interface "${extendDirective.interface}" described in @extend directive for "${type.name}" isn't abstract type or doesn't exist`);
    }
    if (interfaces.includes(extendType)) {
      throw new Error(`The interface "${extendDirective.interface}" described in @extend directive for "${type.name}" is already implemented by the type`);
    }
    interfaces.push(...traverseExtends(extendType, schema));
  }
  return interfaces;
}

const defaultRefToId = (ref) => {
  return typeof ref === "string" ? ref : catalogModel.stringifyEntityRef(ref);
};
function createGraphQLApp(options) {
  const { modules, plugins, loader, refToId = defaultRefToId } = options;
  const application = graphqlModules.createApplication({
    schemaBuilder: ({ typeDefs, resolvers }) => transformDirectives(
      schema.makeExecutableSchema({ typeDefs, resolvers })
    ),
    modules: [Core, Catalog, ...modules != null ? modules : []]
  });
  const run = core.envelop({
    plugins: [
      graphqlModules$1.useGraphQLModules(application),
      dataloader.useDataLoader("loader", loader),
      core.useExtendContext(() => ({ refToId })),
      ...plugins != null ? plugins : []
    ]
  });
  return { run, application };
}

function createLoader(catalog) {
  return new DataLoader__default["default"](async (refs) => {
    const entityRefs = refs.map((ref) => typeof ref === "string" ? ref : catalogModel.stringifyEntityRef(ref));
    const result = await catalog.getEntitiesByRefs({ entityRefs });
    return result.items.map((entity, index) => entity != null ? entity : new core.EnvelopError(`no such node with ref: '${refs[index]}'`));
  });
}

function transformSchema(source) {
  const schema = transformDirectives(
    graphql.buildASTSchema(
      merge.mergeTypeDefs([
        source,
        loadFiles.loadFilesSync(
          backendCommon.resolvePackagePath(
            "@frontside/backstage-plugin-graphql",
            "src/app/modules/**/*.graphql"
          )
        )
      ])
    )
  );
  const errors = graphql.validateSchema(schema);
  if (errors.length > 0) {
    throw new Error(errors.map((e) => e.message).join("\n"));
  }
  return schema;
}

async function createRouter(options) {
  const { logger } = options;
  const { run, application } = createGraphQLApp({
    modules: options.modules,
    loader: () => createLoader(options.catalog),
    plugins: options.plugins
  });
  const router = Router__default["default"]();
  router.use(express__default["default"].json());
  router.use((_, res, next) => {
    res.setHeader("Content-Security-Policy", "'self' http: 'unsafe-inline'");
    next();
  });
  router.get("/schema", (_, response) => {
    response.send(graphql.printSchema(application.schema));
  });
  router.use("/", expressGraphql.graphqlHTTP(async () => {
    const { parse, validate, contextFactory, execute } = run();
    return {
      schema: application.schema,
      graphiql: true,
      customParseFn: parse,
      customValidateFn: validate,
      customExecuteFn: async (args) => {
        return execute({
          ...args,
          contextValue: await contextFactory()
        });
      }
    };
  }));
  router.get("/health", (_, response) => {
    logger.info("PONG!");
    response.send({ status: "ok" });
  });
  router.use(backendCommon.errorHandler());
  return router;
}

exports.createGraphQLApp = createGraphQLApp;
exports.createLoader = createLoader;
exports.createRouter = createRouter;
exports.transformSchema = transformSchema;
//# sourceMappingURL=index.cjs.js.map
